import logging
from threading import Thread

from typing import List
from ._CronTabInteractor import CronTabInteractor


logger = logging.getLogger("SchedComp")


class CronTabService(object):
    def __init__(self, interactors):
        # type: (List[CronTabInteractor]) -> None
        self.interactors = interactors
        self.threads = []
        self.running = False

        for i in range(0, len(self.interactors)):
            t = Thread(target=self.run, args=(i, ))
            t.daemon = True
            self.threads.append(t)

    def start(self):
        self.running = True
        for thread in self.threads:
            thread.start()

    def stop(self):
        self.running = False

    def run(self, index):
        while self.running:
            try:
                self.interactors[index].execute()
            except:
                logger.exception("Error handling cron job: {}".format(self.interactors[index].event_definition))



from typing import Optional, Dict, List


class EventDefinition(object):
    def __init__(self, subject, event_type=None, data=None, headers=None):
        # type: (str, Optional[str], Optional[bytes], List[Dict[str, str]]) -> None
        self.subject = subject
        self.event_type = event_type
        self.data = data
        self.headers = headers



import logging
from datetime import (
    timedelta,
)
from typing import Dict

import iso8601
import redis
import redis_lock
from iso8601 import ParseError
from timeutil import Clock
from ._RedisClusterLock import RedisClusterLock

logger = logging.getLogger(__name__)


class RedisLocker(object):

    key_template = "schedule.event.{}:lastPublished"

    lock_key_template = "schedule.event.{}:lastPublishedLock"

    def __init__(
        self,
        redis_client: redis.StrictRedis,
        clock: Clock,
    ):
        self.redis_client = redis_client
        self.clock = clock

        self.locks: Dict[str, redis_lock.Lock] = {}

    def must_publish(self, subject: str, await_time: float) -> bool:
        lock = self._get_lock(subject=subject)
        acquired = lock.acquire(blocking=False)
        if not acquired:
            return False

        try:
            key = self._get_key(subject=subject)
            next_publish_value = self.redis_client.get(name=key)
            if next_publish_value is None:
                self._mark_next_publish(
                    key=key,
                    await_time=await_time,
                )
                return True

            try:
                next_publish_str = next_publish_value.decode("utf-8")
                next_published = iso8601.parse_date(next_publish_str)
                now = self.clock.now()
                if next_published > now.replace(tzinfo=iso8601.UTC):
                    return False

                self._mark_next_publish(
                    key=key,
                    await_time=await_time,
                )
                return True
            except (UnicodeDecodeError, ParseError):
                self.redis_client.delete(key)
        finally:
            lock.release()

    def _get_key(self, subject: str) -> str:
        return self.key_template.format(subject)

    def _get_lock(self, subject: str) -> redis_lock.Lock:
        if subject not in self.locks:
            self.locks[subject] = RedisClusterLock(
                redis_client=self.redis_client,
                name=self.lock_key_template.format(subject),
                expire=5,
            )

        return self.locks[subject]

    def _mark_next_publish(
        self,
        key: str,
        await_time: float,
    ) -> None:
        next_published = self.clock.now() + timedelta(seconds=await_time - 5)
        logger.info("Next publish: %s", next_published.isoformat())
        self.redis_client.set(name=key, value=next_published.isoformat())



from ._CronTabInteractor import CronTabInteractor
from ._CronTabService import CronTabService
from ._EventDefinition import EventDefinition
from ._InvalidCronTabExpression import InvalidCronTabExpression
from ._RedisLocker import RedisLocker



from redis_lock import Lock


class RedisClusterLock(Lock):
  def __init__(self, redis_client, name, *args, **kwargs):
    super(RedisClusterLock, self).__init__(redis_client, name, *args, **kwargs)
    self._name = f"lock:{{{name}}}"
    self._signal = f"lock-signal:{{{name}}}"



import json
import logging
from hashlib import md5
from typing import (
    Dict,
    List,
)

from crontab import CronTab
from messagebus import (
    Event,
    MessageBus,
)
from timeutil import (
    Clock,
    Sleeper,
)
from ._EventDefinition import EventDefinition
from ._InvalidCronTabExpression import InvalidCronTabExpression
from ._RedisLocker import RedisLocker

logger = logging.getLogger(__name__)


class CronTabInteractor(object):
    def __init__(
        self,
        message_bus: MessageBus,
        expression: str,
        event_definition: EventDefinition,
        clock: Clock,
        sleeper: Sleeper,
        locker: RedisLocker = None,
    ) -> None:

        self.message_bus = message_bus
        self.expression = expression
        self.event_definition = event_definition
        self.clock = clock
        self.sleeper = sleeper
        self.locker = locker

        try:
            self.entry = CronTab(expression)
        except ValueError:
            raise InvalidCronTabExpression(expression)

        self.headers_hash = self._calculate_headers_hash(headers=event_definition.headers)

    def execute(self):
        sleep_time = self.entry.next(self.clock.now(), default_utc=False)
        self.sleeper.sleep(sleep_time)

        if self.locker is not None:
            if not self.locker.must_publish(
                subject=f"{self.event_definition.subject}-{self.expression}-{self.headers_hash}",
                await_time=self.entry.next(self.clock.now(), default_utc=False),
            ):
                logger.debug("Skipping publication of: {}".format(self.event_definition.subject))
                return

        logger.info("Publishing event: {}".format(self.event_definition.subject))
        for headers in self.event_definition.headers:
            logger.info("Publishing event: {} for context {}".format(self.event_definition.subject, headers))
            self.message_bus.publish_event(self.build_event(headers))

    def build_event(self, headers: Dict[str, str]):
        return Event(
            subject=self.event_definition.subject,
            event_type=self.event_definition.event_type,
            data=self.event_definition.data,
            headers=headers,
        )

    @staticmethod
    def _calculate_headers_hash(headers: List[Dict[str, str]]) -> str:
        headers_str = json.dumps(headers)
        return md5(headers_str.encode('utf-8')).hexdigest()



from messagehandler import MessageHandlerBuilder


class SchedCompMessageHandlerBuilder(MessageHandlerBuilder):
    def build_singletons(self):
        pass

    def build_message_handler(self, message, event):
        pass

    def destroy_message_handler(self, message_handler):
        pass

    def destroy_singletons(self):
        pass



import logging
import os
from json import JSONDecodeError

import helper
import zmq
from messagehandlerbuilder import SchedCompMessageHandlerBuilder
from redis import StrictRedis
from schedcomp import (
    CronTabInteractor,
    CronTabService,
    EventDefinition,
    RedisLocker,
)
from timeutil import (
    RealClock,
    RealSleeper,
)
from zeromqmessagehandler import (
    ZeroMqMessageBus,
    ZeroMqMessageHandler,
)

COMP_NAME = "SchedComp"
logger = logging.getLogger(COMP_NAME)


def main():
    config_path = './config.json'
    if "CONFIG_PATH" in os.environ:
        config_path = os.environ["CONFIG_PATH"]
    if os.path.isdir(config_path):
        configs_files = []
        default_config = os.environ.get("DEFAULT_CONFIG_FILE", "config.json")
        complete_default_config_path = os.path.join(config_path, default_config)
        configs_files.append(complete_default_config_path)
        for file in os.listdir(config_path):
            complete_config_path = os.path.join(config_path, file)
            if os.path.isfile(complete_config_path) and \
                    file.endswith(".json") and \
                    complete_config_path != complete_default_config_path:
                configs_files.append(complete_config_path)
    else:
        configs_files = config_path.split(",")

    config = helper.read_json_from_file_as_dictionary(configs_files[0])
    if "cronTab" not in config:
        config["cronTab"] = []
    for crontab_file in configs_files[1:]:
        try:
            file_crontab = helper.read_json_from_file_as_dictionary(crontab_file)
            if not isinstance(file_crontab, list):
                raise Exception("The crontab file must be a list of event definitions: {}".format(crontab_file))
        except JSONDecodeError as ex:
            raise Exception("Error parsing config file {}".format(crontab_file)) from ex
        config["cronTab"].extend(file_crontab)

    log_path = config_path
    if "LOG_PATH" in os.environ:
        log_path = os.environ["LOG_PATH"]
    helper.config_logger(log_path, COMP_NAME, add_error_handler=True)

    logger.debug("Configuration file:\n{}".format(config))

    context = zmq.Context()
    message_bus = ZeroMqMessageBus(COMP_NAME, context)
    clock = RealClock()
    sleeper = RealSleeper()
    interactors = []

    locker = None
    if "redis" in config:
        locker = RedisLocker(
            redis_client=StrictRedis(
                host=config["redis"]["host"],
                port=config["redis"]["port"],
                db=config["redis"]["db"],
                socket_timeout=2,
                socket_connect_timeout=2,
                ssl=config["redis"].get("ssl", False),
            ),
            clock=clock,
        )

    for entry in config["cronTab"]:
        event_definition = EventDefinition(
            entry["subject"],
            entry.get("type"),
            entry.get("data"),
            entry["headers"]
        )

        interactors.append(
            CronTabInteractor(
                message_bus=message_bus,
                expression=entry["expression"],
                event_definition=event_definition,
                clock=clock,
                sleeper=sleeper,
                locker=locker,
            )
        )

    cron_tab_service = CronTabService(interactors)
    cron_tab_service.start()

    message_handler = ZeroMqMessageHandler(context,
                                           COMP_NAME,
                                           message_bus,
                                           SchedCompMessageHandlerBuilder())
    message_handler.handle_messages()


if __name__ == '__main__':
    main()
